import logging
import os
import litellm

import asyncio
import json
import time

from typing import Union, Dict, Any, Optional, List, TypedDict
from openai import OpenAIError
from litellm.utils import ModelResponse, CustomStreamWrapper

from xgae.utils import to_bool
from xgae.utils.setup_env import setup_langfuse

class LLMConfig(TypedDict, total=False):
    model: str              # Optional Name of the model to use , Override .env LLM_MODEL
    model_name: str         # Optional Name of the model to use , use model if empty
    model_id: str           # Optional ARN for Bedrock inference profiles, default is None
    api_key: str            # Optional  API key, Override .env LLM_API_KEY or OS env variable
    api_base: str           # Optional API base URL, Override .env LLM_API_BASE
    temperature: float      # temperature: Optional Sampling temperature (0-1), Override .env LLM_TEMPERATURE
    max_tokens: int         # max_tokens: Optional Maximum tokens in the response, Override .env LLM_MAX_TOKENS
    stream: bool            # stream: Optional whether to stream the response, Override .env LLM_STREAM
    enable_thinking: bool   # Optional whether to enable thinking, Override .env LLM_ENABLE_THINKING
    reasoning_effort: str   # Optional level of reasoning effort, default is  ‘low’
    response_format: str    # response_format: Optional desired format for the response, default is  None
    top_p: int              # Optional Top-p sampling parameter, default is None


class LangfuseMetadata(TypedDict, total=False):
    generation_name: str
    generation_id: str
    existing_trace_id: str
    session_id: str


class LLMError(Exception):
    """Base exception for LLM-related errors."""
    pass

class LLMClient:
    RATE_LIMIT_DELAY = 30
    RETRY_DELAY = 0.1

    langfuse_inited = False
    langfuse_enabled = False

    def __init__(self, llm_config: LLMConfig=None):
        litellm.modify_params = True
        litellm.drop_params = True
        litellm.disable_token_counter = True
        litellm.disable_end_user_cost_tracking = True
        litellm.disable_end_user_cost_tracking_prometheus_only = True

        self._init_langfuse()

        llm_config = llm_config or LLMConfig()
        if llm_config.get('model') and llm_config.get('model_name') is None:
            llm_config['model_name'] = llm_config.get('model')

        self.max_retries = int(os.getenv('LLM_MAX_RETRIES', 1))

        env_llm_model           = os.getenv('LLM_MODEL', "openai/qwen3-235b-a22b")
        env_llm_api_key         = os.getenv('LLM_API_KEY')
        env_llm_api_base        = os.getenv('LLM_API_BASE', "https://dashscope.aliyuncs.com/compatible-mode/v1")
        env_llm_max_tokens      = int(os.getenv('LLM_MAX_TOKENS', 8192))
        env_llm_temperature     = float(os.getenv('LLM_TEMPERATURE', 0.7))
        env_llm_stream          = to_bool(os.getenv('LLM_STREAM', False))
        env_llm_enable_thinking = to_bool(os.getenv('LLM_ENABLE_THINKING', False))

        llm_config_params = {
            'model':            llm_config.get('model', env_llm_model),
            'model_name':       llm_config.get('model_name', env_llm_model),
            'model_id':         llm_config.get('model_id', None),
            'api_key':          llm_config.get('api_key', env_llm_api_key),
            'api_base':         llm_config.get('api_base', env_llm_api_base),
            'temperature':      llm_config.get('temperature', env_llm_temperature),
            'max_tokens':       llm_config.get('max_tokens', env_llm_max_tokens),
            'stream':           llm_config.get('stream', env_llm_stream),
            'enable_thinking':  llm_config.get('enable_thinking', env_llm_enable_thinking),
            'reasoning_effort': llm_config.get('reasoning_effort', "low"),
            'response_format':  llm_config.get('response_format', None),
            'top_p':            llm_config.get('top_p', None),
            'tools':            None,
            'tool_choice':      None,
        }

        self.model_name = llm_config_params['model_name']
        self.is_stream = llm_config_params['stream']

        self.lite_llm_params = self._prepare_llm_params(llm_config_params)
        logging.info(f"=== LLMClient initialed : model={self.model_name}, is_stream={self.is_stream}, enable thinking={self.lite_llm_params['enable_thinking']}")

    @staticmethod
    def _init_langfuse():
        if not LLMClient.langfuse_inited:
            LLMClient.langfuse_inited =True

            env_llm_langfuse_enable = to_bool(os.getenv("LLM_LANGFUSE_ENABLE", False))
            if env_llm_langfuse_enable:
                env_langfuse = setup_langfuse()
                if env_langfuse and env_langfuse.enabled:
                    litellm.success_callback = ["langfuse"]
                    litellm.failure_callback = ["langfuse"]
                    LLMClient.langfuse_enabled = True
                    logging.info("🛠️ LiteLLM Langfuse is enable !")
                else:
                    LLMClient.langfuse_enabled = False
                    logging.warning("🛠️ LiteLLM Langfuse is disable, langfuse.enabled=false !")
            else:
                LLMClient.langfuse_enabled = False
                logging.warning("🛠️ LiteLLM Langfuse is disable, LLM_LANGFUSE_ENABLE=False !")

    def _prepare_llm_params(self, llm_config_params: Dict[str, Any]) -> Dict[str, Any]:
        prepared_llm_params = llm_config_params.copy()

        model_name = llm_config_params.get("model_name")
        max_tokens = llm_config_params.get("max_tokens")
        model_id = llm_config_params.get("model_id")

        # Handle token limits
        if max_tokens is not None:
            # For Claude 3.7 in Bedrock, do not set max_tokens or max_tokens_to_sample
            # as it causes errors with inference profiles
            if model_name.startswith("bedrock/") and "claude-3-7" in model_name:
                prepared_llm_params.pop("max_tokens")
                logging.debug(f"LLMClient prepare_llm_params: Remove 'max_tokens' param for model: {model_name}")
            else:
                is_openai_o_series = 'o1' in model_name
                is_openai_gpt5 = 'gpt-5' in model_name
                param_name = "max_completion_tokens" if (is_openai_o_series or is_openai_gpt5) else "max_tokens"
                if param_name == "max_completion_tokens":
                    prepared_llm_params[param_name] = max_tokens
                    logging.debug(f"LLMClient prepare_llm_params: Add 'max_completion_tokens' param for model: {model_name}")

        # # Add Claude-specific headers
        if "claude" in model_name.lower() or "anthropic" in model_name.lower():
            prepared_llm_params["extra_headers"] = {
                "anthropic-beta": "output-128k-2025-02-19"
            }
            logging.debug(f"LLMClient prepare_llm_params: Add 'extra_headers' param for model: {model_name}")

        # Add Bedrock-specific parameters
        if model_name.startswith("bedrock/"):
            if not model_id and "anthropic.claude-3-7-sonnet" in model_name:
                prepared_llm_params["model_id"] = "arn:aws:bedrock:us-west-2:935064898258:inference-profile/us.anthropic.claude-3-7-sonnet-20250219-v1:0"
                logging.debug(f"LLMClient prepare_llm_params: Must Set 'model_id' param for model: {model_name}")

        # Apply Anthropic prompt caching (minimal implementation)
        effective_model_name = llm_config_params.get("model", model_name)

        # OpenAI GPT-5: drop unsupported temperature param (only default 1 allowed)
        if "gpt-5" in effective_model_name and "temperature" in llm_config_params and llm_config_params["temperature"] != 1:
            prepared_llm_params.pop("temperature", None)
            logging.debug(f"LLMClient prepare_llm_params: Remove 'temperature' param for model: {model_name}")

        # OpenAI GPT-5: request priority service tier when calling OpenAI directly
        # Pass via both top-level and extra_body for LiteLLM compatibility
        if "gpt-5" in effective_model_name and not effective_model_name.startswith("openrouter/"):
            prepared_llm_params["service_tier"] = "priority"
            prepared_llm_params["extra_body"] = {"service_tier": "priority"}
            logging.debug(f"LLMClient prepare_llm_params: Add 'service_tier' and 'extra_body' param for model: {model_name}")

        # Add reasoning_effort for Anthropic models if enabled
        enable_thinking = llm_config_params.get("enable_thinking")
        use_thinking = enable_thinking if enable_thinking is not None else False

        is_anthropic = "anthropic" in effective_model_name.lower() or "claude" in effective_model_name.lower()
        is_kimi_k2 = "kimi-k2" in effective_model_name.lower() or model_name.startswith("moonshotai/kimi-k2")

        if is_kimi_k2:
            prepared_llm_params["provider"] = {
                "order": ["together/fp8", "novita/fp8", "baseten/fp8", "moonshotai", "groq"]
            }
            logging.debug(f"LLMClient prepare_llm_params: Add 'provider' param for model: {model_name}")

        reasoning_effort = llm_config_params.get("reasoning_effort")
        if is_anthropic and use_thinking:
            effort_level = reasoning_effort if reasoning_effort else 'low'
            prepared_llm_params["reasoning_effort"] = effort_level
            prepared_llm_params["temperature"] = 1.0  # Required by Anthropic when reasoning_effort is used
            logging.debug(f"LLMClient prepare_llm_params: Set 'temperature'=1.0 param for model: {model_name}")

        return prepared_llm_params


    def _prepare_complete_params(self, messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Prepare parameters for the API call."""
        complete_params = self.lite_llm_params.copy()
        complete_params["messages"] = messages

        model_name = self.lite_llm_params["model_name"]
        effective_model_name = complete_params.get("model", model_name)

        # Apply cache control to the first 4 text blocks across all messages , for anthropic and claude model
        if "claude" in effective_model_name.lower() or "anthropic" in effective_model_name.lower():
            messages = complete_params["messages"]

            if not isinstance(messages, list):
                return complete_params

            cache_control_count = 0
            max_cache_control_blocks = 3

            for message in messages:
                if cache_control_count >= max_cache_control_blocks:
                    break

                content = message.get("content")

                if isinstance(content, str):
                    message["content"] = [
                        {"type": "text", "text": content, "cache_control": {"type": "ephemeral"}}
                    ]
                    cache_control_count += 1
                    logging.debug(f"LLMClient prepare_complete_params: Add 'cache_control' in message content, for model: {model_name}")
                elif isinstance(content, list):
                    for item in content:
                        if cache_control_count >= max_cache_control_blocks:
                            break
                        if isinstance(item, dict) and item.get("type") == "text" and "cache_control" not in item:
                            item["cache_control"] = {"type": "ephemeral"}
                            cache_control_count += 1
                            logging.debug(f"LLMClient prepare_complete_params: Add 'cache_control' in message content list, for model: {model_name}")

        return complete_params


    async def acompletion(self,
                          messages: List[Dict[str, Any]],
                          langfuse_metadata: Optional[LangfuseMetadata]=None
                          ) -> Union[ModelResponse, CustomStreamWrapper]:
        complete_params = self._prepare_complete_params(messages)
        if LLMClient.langfuse_enabled and langfuse_metadata:
            complete_params["metadata"] = langfuse_metadata

        last_error = None
        for attempt in range(self.max_retries):
            try:
                logging.info(f"*** LLMClient acompletion: LLM '{self.model_name}' completion attempt {attempt + 1}/{self.max_retries}")
                response = await litellm.acompletion(**complete_params)
                return response
            except (litellm.exceptions.RateLimitError, OpenAIError, json.JSONDecodeError) as error:
                last_error = error
                if (attempt + 1) < self.max_retries:
                    delay = LLMClient.RATE_LIMIT_DELAY if isinstance(error, litellm.exceptions.RateLimitError) else LLMClient.RETRY_DELAY
                    logging.warning(f"LLMClient acompletion: retry={attempt+1}/{self.max_retries}, delay={delay}, error:"
                                    f"\n {error}")
                    await asyncio.sleep(delay)
            except Exception as e:
                logging.error(f"LLMClient acompletion: Unexpected error during LLM completion: {str(e)}", exc_info=True)
                raise LLMError(f"LLMClient create completion failed: {e}")

        logging.error(f"LLMClient acompletion: LLM completion failed after {self.max_retries} attempts: {last_error}", exc_info=True)
        raise LLMError(f"LLMClient create completion failed after {self.max_retries} attempts !")


    def completion(self,
                   messages: List[Dict[str, Any]],
                   langfuse_metadata: Optional[LangfuseMetadata] = None
                   ) -> Union[ModelResponse, CustomStreamWrapper]:
        complete_params = self._prepare_complete_params(messages)
        if LLMClient.langfuse_enabled and langfuse_metadata:
            complete_params["metadata"] = langfuse_metadata

        last_error = None

        for attempt in range(self.max_retries):
            try:
                logging.info(f"*** LLMClient completion: LLM '{self.model_name}' completion attempt {attempt + 1}/{self.max_retries}")
                response = litellm.completion(**complete_params)
                return response
            except (litellm.exceptions.RateLimitError, OpenAIError, json.JSONDecodeError) as error:
                last_error = error
                if (attempt + 1) < self.max_retries:
                    delay = LLMClient.RATE_LIMIT_DELAY if isinstance(error, litellm.exceptions.RateLimitError) else LLMClient.RETRY_DELAY
                    logging.warning(f"LLMClient completion: retry={attempt+1}/{self.max_retries}, delay={delay}, error:"
                                    f"\n {error}")
                    time.sleep(delay)
            except Exception as e:
                logging.error(f"LLMClient completion: Unexpected error during LLM completion: {str(e)}", exc_info=True)
                raise LLMError(f"LLMClient create completion failed: {e}")

        logging.error(f"LLMClient completion: LLM completion failed after {self.max_retries} attempts: {last_error}", exc_info=True)
        raise LLMError(f"LLMClient create completion failed after {self.max_retries} attempts !")


    async def get_acompletion_response(self, response: Union[ModelResponse, CustomStreamWrapper]) -> str:
        response_text: str = ""

        if self.is_stream:
            async for chunk in response:
                choices = chunk.get("choices", [{}])
                if not choices:
                    continue
                delta = choices[0].get("delta", {})
                content = delta.get("content", "")
                if content:
                    response_text += content
        else:
            response_text = response.choices[0].message.content

        return response_text


    def get_completion_response(self, response: Union[ModelResponse, CustomStreamWrapper]) -> str:
        response_text: str = ""

        if self.is_stream:
            for chunk in response:
                choices = chunk.get("choices", [{}])
                if not choices:
                    continue
                delta = choices[0].get("delta", {})
                content = delta.get("content", "")
                if content:
                    response_text += content
        else:
            response_text = response.choices[0].message.content

        return response_text



if __name__ == "__main__":
    from xgae.utils.setup_env import setup_logging

    setup_logging()
    langfuse = setup_langfuse()
    llm_client = LLMClient(LLMConfig(stream=True))

    async def async_call_llm():
        messages = [{"role": "user", "content": "1+1="}]
        trace_id = langfuse.trace(name = "xgae_acompletion_test").trace_id
        await asyncio.sleep(1)

        meta = LangfuseMetadata(
                generation_name="llm_acompletion_test",
                generation_id="generation_id_0",
                existing_trace_id=trace_id,
                session_id="session_0",
        )

        response = await llm_client.acompletion(messages, meta)
        result = await llm_client.get_acompletion_response(response)
        print(f"acompletion result={result}")

    def call_llm():
        messages = [{"role": "user", "content": "2+2="}]
        trace_id = langfuse.trace(name = "xgae_completion_test").trace_id
        time.sleep(1)

        meta = LangfuseMetadata(
                generation_name="llm_completion_test",
                generation_id="generation_id_1",
                existing_trace_id=trace_id,
                session_id="session_1",
        )

        response = llm_client.completion(messages, meta)
        result = llm_client.get_completion_response(response)
        print(f"completion result={result}")


    print("==========   Begin Sync LLM Call Test   =================")
    call_llm()

    print("==========   Begin Aysnc LLM Call Test   =================")
    asyncio.run(async_call_llm())





"""
JSON helper utilities for handling both legacy (string) and new (dict/list) formats.

These utilities help with the transition from storing JSON as strings to storing
them as proper JSONB objects in the database.
"""

import json
from typing import Any, Union, Dict, List


def ensure_dict(value: Union[str, Dict[str, Any], None], default: Dict[str, Any] = None) -> Dict[str, Any]:
    """
    Ensure a value is a dictionary.
    
    Handles:
    - None -> returns default or {}
    - Dict -> returns as-is
    - JSON string -> parses and returns dict
    - Other -> returns default or {}
    
    Args:
        value: The value to ensure is a dict
        default: Default value if conversion fails
        
    Returns:
        A dictionary
    """
    default = default or {}
        
    if value is None:
        return default
        
    if isinstance(value, dict):
        return value
        
    if isinstance(value, str):
        try:
            parsed = json.loads(value)
            if isinstance(parsed, dict):
                return parsed
            return default
        except (json.JSONDecodeError, TypeError):
            return default
            
    return default


def ensure_list(value: Union[str, List[Any], None], default: List[Any] = None) -> List[Any]:
    """
    Ensure a value is a list.
    
    Handles:
    - None -> returns default or []
    - List -> returns as-is
    - JSON string -> parses and returns list
    - Other -> returns default or []
    
    Args:
        value: The value to ensure is a list
        default: Default value if conversion fails
        
    Returns:
        A list
    """
    default = default or []
        
    if value is None:
        return default
        
    if isinstance(value, list):
        return value
        
    if isinstance(value, str):
        try:
            parsed = json.loads(value)
            if isinstance(parsed, list):
                return parsed
            return default
        except (json.JSONDecodeError, TypeError):
            return default
            
    return default

# @todo if all call value is str, delete useless code
def safe_json_parse(value: Union[str, Dict, List, Any], default: Any = None) -> Any:
    """
    Safely parse a value that might be JSON string or already parsed.
    
    This handles the transition period where some data might be stored as
    JSON strings (old format) and some as proper objects (new format).
    
    Args:
        value: The value to parse
        default: Default value if parsing fails
        
    Returns:
        Parsed value or default
    """
    if value is None:
        return default
        
    # If it's already a dict or list, return as-is
    if isinstance(value, (dict, list)):
        return value

    if isinstance(value, str):
        try:
            return json.loads(value)
        except (json.JSONDecodeError, TypeError):
            return value

    return value


def to_json_string(value: Any) -> str:
    """
    Convert a value to a JSON string if needed.
    
    This is used for backwards compatibility when yielding data that
    expects JSON strings.
    
    Args:
        value: The value to convert
        
    Returns:
        JSON string representation
    """
    if isinstance(value, str):
        # If it's already a string, check if it's valid JSON
        try:
            json.loads(value)
            return value  # It's already a JSON string
        except (json.JSONDecodeError, TypeError):
            pass

    # For all other types, convert to JSON
    return json.dumps(value)




from functools import partial

from are.simulation.agents.llm.llm_engine import LLMEngine
from are.simulation.agents.are_simulation_agent_config import ARESimulationReactBaseAgentConfig
from are.simulation.agents.default_agent.steps.are_simulation import get_are_simulation_update_pre_step
from are.simulation.agents.default_agent.termination_methods.are_simulation import get_gaia2_termination_step
from are.simulation.agents.default_agent.tools.json_action_executor import JsonActionExecutor
from are.simulation.agents.default_agent.base_agent import TerminationStep
from are.simulation.agents.default_agent.termination_methods.are_simulation import  termination_condition_are_simulation

from are.simulation.agents.xga.are_agent import XGAAreAgent

termination_condition_are_simulation_send_message_to_user = partial(
    termination_condition_are_simulation,
    termination_tool_names=[
        "AgentUserInterface__send_message_to_user"
    ],
)

def get_xga_termination_step() -> TerminationStep:
    return TerminationStep(
        name="end_scenario",
        condition=termination_condition_are_simulation_send_message_to_user,
    )

def xga_simulation_react_xml_agent(
    llm_engine: LLMEngine, base_agent_config: ARESimulationReactBaseAgentConfig
):
    return XGAAreAgent(
        llm_engine=llm_engine,
        tools={},
        system_prompts={
            "system_prompt": str(base_agent_config.system_prompt),
        },
        termination_step=get_xga_termination_step(),
        max_iterations=base_agent_config.max_iterations,
        action_executor=JsonActionExecutor(  # Just for compatible BaseAgent， useless
            use_custom_logger=base_agent_config.use_custom_logger
        ),
        conditional_pre_steps=[get_are_simulation_update_pre_step()],
        use_custom_logger=base_agent_config.use_custom_logger,
    )



import json
import logging

from typing import Any, Dict, Optional
from typing_extensions import override

from langchain_mcp_adapters.tools import load_mcp_tools

from xgae.engine.engine_base import XGAError, XGAToolSchema, XGAToolResult, XGAToolType
from xgae.engine.mcp_tool_box import XGAMcpToolBox

from are.simulation.tools import Tool
from are.simulation.agents.xga.mcp_tool_executor import convert_mcp_tool_result

class XGAAreToolBox(XGAMcpToolBox):
    def __init__(self, are_tools: Dict[str, Tool]):
        super().__init__()
        self.are_tools = are_tools
        self._is_loaded_are_tool_schemas = False


    @override
    async def init_tool_schemas(self):
        #await self._load_mcp_tools_schema()
        self._load_are_tools_schema()


    @override
    async def call_tool(self, task_id: str, tool_name: str, args: Optional[Dict[str, Any]] = None) -> XGAToolResult:
        task_tool_schemas = self.task_tool_schemas.get(task_id, {})
        tool_schema = task_tool_schemas.get(tool_name, None)
        if tool_schema is None:
            raise XGAError(f"ARE tool not found: '{tool_name}'")
        server_name = tool_schema.server_name

        tool_type = self._get_tool_type(server_name)

        if tool_type == "system":
            tool_args = args or {}
            tool_args = dict({'task_id': task_id}, **tool_args)
            result = self.system_tools.call_tool(tool_name, tool_args)
        elif tool_type == "custom":  # ARE Tools
            try:
                tool_result = self.are_tools[tool_name](**args)
                result = convert_mcp_tool_result(str(tool_result))
            except Exception as e:
                error = f"Call ARE Tool '{tool_name}' error: {str(e)}"
                logging.error(f"AreToolBox call_are_tool: {error}")
                result = XGAToolResult(success=False, output=str(e))
        else:
            raise XGAError(f"ARE tool no support type: '{tool_type}'")

        return result


    async def reload_mcp_tools_schema(self) -> None:
        self._is_loaded_mcp_tool_schemas = False
        self._is_loaded_are_tool_schemas = False
        await self.init_tool_schemas()


    def _load_are_tools_schema(self) -> None:
        if not self._is_loaded_are_tool_schemas:
            for are_tool in self.are_tools.values():
                tool_name = are_tool.name
                server_name , unction_name = tool_name.split("__")
                tool_type :XGAToolType = "custom"
                input_schema = {
                    'properties': are_tool.inputs,
                    'required': []
                }

                tool_schema = XGAToolSchema(tool_name       = tool_name,
                                            function_name   = tool_name,
                                            tool_type       = tool_type,
                                            server_name     = server_name,
                                            description     = are_tool.description,
                                            input_schema    = input_schema,
                                            metadata        = {}
                                            )
                if server_name not in self.mcp_tool_schemas:
                    self.mcp_tool_schemas[server_name] = []
                self.mcp_tool_schemas[server_name].append(tool_schema)
                if server_name not in self.mcp_server_names:
                    self.mcp_server_names.append(server_name)

        self._is_loaded_are_tool_schemas = True


import click
import logging

from typing import Annotated, Dict, Any
from pydantic import Field

from mcp.server.fastmcp import FastMCP

mcp = FastMCP(name="Fault Location Tools")
alarm_type = 0

@mcp.tool(
    description="Get Alarm Object",
)
def get_alarm(ip: Annotated[str, Field(description="Alarm Object IP Address")]) -> Dict[str, str]:
    logging.info(f"get_alarm: ip={ip}")
    return {"alarmId":"alm0123", "objId": "obj567"}


@mcp.tool(
    description="Get alarm type, return result enum: 1: Business Type Alarm; 2: Equipment Type Alarm; 3: Middleware Type Alarm",
)
def get_alarm_type(alarm_id: Annotated[str, Field(description="Alarm Object Id")]) -> str:
    logging.info(f"get_alarm_type: alarm_id={alarm_id}, alarm_type={alarm_type}")
    return f"{alarm_type}"


@mcp.tool(
    description="Locate Business Type Fault",
)
def get_busi_fault(obj_id: Annotated[str, Field(description="Alarm Object Id")]) -> Dict[str, Any]:
    logging.info(f"get_busi_fault: obj_id={obj_id}")
    return {"fault_code": "F01", "fault_desc": "Business Recharge Fault"}


@mcp.tool(
    description="Locate Equipment Type Fault",
)
def get_equip_fault(obj_id: Annotated[str, Field(description="Alarm Object Id")]) -> Dict[str, Any]:
    logging.info(f"get_equip_fault: obj_id={obj_id}")
    return {"fault_code": "F02", "fault_desc": "Host Disk Fault"}


@mcp.tool(
    description="Locate Middleware Type Fault",
)
def get_middle_fault(obj_id: Annotated[str, Field(description="Alarm Object Id")]) -> Dict[str, Any]:
    logging.info(f"get_middle_fault: obj_id={obj_id}")
    return {"fault_code": "F03", "fault_desc": "Redis Fault"}


@mcp.tool(
    description="Get Business Type Fault Solution and Cause",
)
async def get_busi_fault_cause(fault_code: Annotated[str, Field(description="Fault Code")]) -> str:
    logging.info(f"get_busi_fault_cause: faultCode={fault_code}")

    fault_cause = ""
    if fault_code == 'F01':
        fault_cause = "Business Recharge Fault, Fault Cause is 'Phone Recharge Application Crash' ，Solution is 'Restart Phone Recharge Application'"
    elif fault_code == 'F02':
        fault_cause = f"FaultCode '{fault_code}' is not Business Type, can use equip related fault cause tool"
    else:
        fault_cause = f"FaultCode '{fault_code}' is not Business Type, can use web_search tool find fault cause"

    return  fault_cause


# @mcp.tool(
#     description="Get Equipment Type Fault Solution and Cause",
# )
# async def query_equip_fault_cause(fault_code: Annotated[str, Field(description="Fault Code")]) -> str:
#     logging.info(f"get_equip_fault_cause: faultCode={fault_code}")
#
#     fault_cause = ""
#     if (fault_code == 'F02'):
#         fault_cause = "Host Fault, Fault Cause is 'Host Disk is Damaged' ，Solution is 'Change Host Disk'"
#     else:
#         fault_cause = f"FaultCode '{fault_code}' is not Equipment Type"
#
#     return  fault_cause


@click.command()
@click.option("--transport", type=click.Choice(["stdio", "sse"]), default="sse", help="Transport type")
@click.option("--host", default="0.0.0.0", help="Host to listen on for SSE")
@click.option("--port", default=17070, help="Port to listen on for SSE")
@click.option("--alarmtype", default=1, help="AlarmType Set")
def main(transport: str, host: str, port: int, alarmtype:int):
    if transport != "stdio":
        from xgae.utils.setup_env import setup_logging
        setup_logging()
        logging.info("=" * 10 + f"   Example Fault Tools Sever Started   " + "=" * 10)
        logging.info(f"=== transport={transport}, host={host}, port={port}, alarmtype={alarmtype}")

    global alarm_type
    alarm_type = alarmtype

    mcp.settings.host = host
    mcp.settings.port = port

    mcp.run(transport=transport)


if __name__ == '__main__':
    main()



import asyncio

from xgae.engine.task_engine import XGATaskEngine
from xgae.utils.setup_env import setup_logging
from xgae.engine.mcp_tool_box import XGAMcpToolBox

setup_logging()

tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")

# Before run examples execute with 'xgatools' project: uv run xgatools

async def run_only_general_tools() -> None:
    engine = XGATaskEngine(tool_box=tool_box, general_tools=["*"], custom_tools=[])

    print("Use only general tools")
    user_input =  "Give me this week's beijing weather, and output with markdown table format"

    final_result = await engine.run_task_with_final_answer(task_input={"role": "user", "content": user_input})
    print("FINAL RESULT:", final_result)
    #print(engine.task_prompt)

# Before run this example execute: uv run example-fault-tools --alarmtype 3
async def run_general_plus_custom_tools() -> None:
    print("\n")
    print("*"*100)
    engine = XGATaskEngine(tool_box=tool_box, general_tools=["*"], custom_tools=["*"])
    print("Use external custom tools with general tools, under no custom prompt")

    # This is not a good question, just show general agent plus custom tools mode
    # If you use custom tools mainly, you should give engine 'system_prompt'
    user_input = "get 10.0.0.1 fault cause"
    final_result = await engine.run_task_with_final_answer(task_input={"role": "user", "content": user_input})
    print("FINAL RESULT:", final_result)
    #print(engine.task_prompt)

async def main() -> None:
    await run_only_general_tools()

    #await run_general_plus_custom_tools()

asyncio.run(main())


import logging

from typing import List, Dict, Any, AsyncGenerator,Optional
from typing_extensions import override

from xgae.utils import log_trace


from xgae.engine.responser.responser_base import TaskResponseProcessor, TaskResponserContext, TaskRunContinuousState


class NonStreamTaskResponser(TaskResponseProcessor):
    def __init__(self, response_context: TaskResponserContext):
        super().__init__(response_context)

    @override
    async def process_response(self,
                               llm_response: Any,prompt_messages: List[Dict[str, Any]],
                               continuous_state: TaskRunContinuousState
                               ) -> AsyncGenerator[Dict[str, Any], None]:
        llm_content = ""
        parsed_xml_data = []
        finish_reason = None
        auto_continue_count = continuous_state['auto_continue_count']

        try:
            if hasattr(llm_response, 'choices') and llm_response.choices:
                if hasattr(llm_response.choices[0], 'finish_reason'):
                    finish_reason = llm_response.choices[0].finish_reason # LLM finish reason: ‘stop' , 'length'
                    logging.info(f"NonStreamResp: LLM response finish_reason={finish_reason}")

                response_message = llm_response.choices[0].message if hasattr(llm_response.choices[0], 'message') else None
                if response_message:
                    if hasattr(response_message, 'content') and response_message.content:
                        llm_content = response_message.content
                        
                        parsed_xml_data = self._parse_xml_tool_calls(llm_content)
                        if self.max_xml_tool_calls > 0 and len(parsed_xml_data) > self.max_xml_tool_calls:
                            logging.warning(f"NonStreamResp: Over XML Tool Limit, finish_reason='xml_tool_limit_reached', "
                                            f"parsed_xml_data_len={len(parsed_xml_data)}")
                            parsed_xml_data = parsed_xml_data[:self.max_xml_tool_calls]
                            finish_reason = "xml_tool_limit_reached"

            self.root_span.event(name=f"non_stream_processor_start[{self.task_no}]({auto_continue_count})", level="DEFAULT",
                             status_message=f"finish_reason={finish_reason}, tool_exec_strategy={self.tool_exec_strategy}, "
                                            f"parsed_xml_data_len={len(parsed_xml_data)}, llm_content_len={len(llm_content)}")

            message_data = {"role": "assistant", "content": llm_content}
            assistant_msg = self.add_response_message(type="assistant", content=message_data, is_llm_message=True)
            yield assistant_msg

            tool_calls_to_execute = [item['tool_call'] for item in parsed_xml_data]
            if  len(tool_calls_to_execute) > 0:
                tool_results = await self._execute_tools(tool_calls_to_execute, self.tool_exec_strategy)

                tool_index = 0
                for i, (returned_tool_call, tool_result) in enumerate(tool_results):
                    parsed_xml_item = parsed_xml_data[i]
                    tool_call = parsed_xml_item['tool_call']
                    parsing_details = parsed_xml_item['parsing_details']
                    assistant_msg_id = assistant_msg['message_id'] if assistant_msg else None

                    tool_context = self._create_tool_context(tool_call, tool_index, assistant_msg_id, parsing_details, tool_result)

                    tool_start_msg = self._add_tool_start_message(tool_context)
                    yield tool_start_msg

                    tool_message = self._add_tool_messsage(tool_context, self.xml_adding_strategy)

                    tool_completed_msg = self._add_tool_completed_message(tool_context, tool_message['message_id'])
                    yield tool_completed_msg

                    yield tool_message

                    if tool_context.function_name in ['ask', 'complete']:
                        finish_reason = "completed"
                        break

                    tool_index += 1
            else:
                finish_reason = "non_tool_call"
                logging.warning(f"NonStreamResp: finish_reason='non_tool_call', No Tool need to call !")

            if finish_reason:
                finish_content = {'status_type': "finish", 'finish_reason': finish_reason}
                finish_msg = self.add_response_message(type="status", content=finish_content, is_llm_message=False)
                yield finish_msg
        except Exception as e:
            trace = log_trace(e, f"NonStreamResp: Process response llm_content:\n {llm_content}")
            self.root_span.event(name="non_stream_process_response_error", level="ERROR",
                                 status_message=f"Process non-streaming response error: {e}",
                                 metadata={"content": llm_content, "trace": trace})

            content = {'role': "system", 'status_type': "error", 'message': f"Process non-streaming response error: {e}"}
            error_msg = self.add_response_message(type="status", content=content, is_llm_message=False)
            yield error_msg

            raise  # Use bare 'raise' to preserve the original exception with its traceback






import click
import logging

from typing import Annotated, Optional, Dict, List, Any, Literal, TypedDict
from pydantic import Field

from mcp.server.fastmcp import FastMCP
from xgae.engine.engine_base import XGAToolResult

mcp = FastMCP(name="Simulate A2A MCP Proxy")

class XGAAgentResult(TypedDict, total=False):
    type: Literal["ask", "answer", "error"]
    content: str
    attachments: Optional[List[str]]

@mcp.tool(
    description="Get Equipment Type Fault Solution and Cause",
)
def query_equip_fault_cause(task_id:str, input: Annotated[str, Field(description="Fault Code")]):
    logging.info(f"query_equip_fault_cause: task_id={task_id}, faultCode={input}")

    fault_cause:XGAAgentResult = None
    if 'F02' in input:
        fault_cause:XGAAgentResult = {
            'type': "answer",
            'content': "Host Fault, Fault Cause is 'Host Disk is Damaged' ，Solution is 'Change Host Disk'"
        }
    else:
        fault_cause:XGAAgentResult = {
            'type': "ask",
            'content': f"input your Equip Fault Code"
        }

    return  fault_cause


@mcp.tool()
def end_xgae_task(task_id:str):
    print(f"SIMU A2A TOOLS: end_xgae_task: task_id={task_id}")
    return XGAToolResult(success=True, output="A2A End Task Success")

@click.command()
@click.option("--transport", type=click.Choice(["stdio", "sse"]), default="sse", help="Transport type")
@click.option("--host", default="0.0.0.0", help="Host to listen on for SSE")
@click.option("--port", default=21010, help="Port to listen on for SSE")
def main(transport: str, host: str, port: int):
    if transport != "stdio":
        from xgae.utils.setup_env import setup_logging
        setup_logging()
        logging.info("=" * 10 + f"   Simulate A2A MCP Proxy Sever Started   " + "=" * 10)
        logging.info(f"=== transport={transport}, host={host}, port={port}")


    mcp.settings.host = host
    mcp.settings.port = port

    mcp.run(transport=transport)


if __name__ == '__main__':
    main()



import logging
from typing import Any, Dict, Optional
from langfuse import Langfuse

from xgae.utils.setup_env import setup_langfuse
from xgae.utils.llm_client import LangfuseMetadata
from xgae.engine.engine_base import XGATaskResult

class XGATaskLangFuse:
    langfuse: Langfuse = None

    def __init__(self,
                 task_id:str,
                 task_run_id: str,
                 task_no: int,
                 session_id: str,
                 user_id: str,
                 agent_id: str) -> None:
        if XGATaskLangFuse.langfuse is None:
            XGATaskLangFuse.langfuse =  setup_langfuse()

        self.session_id     = session_id
        self.task_id        = task_id
        self.task_run_id    = task_run_id
        self.task_no        = task_no
        self.user_id        = user_id
        self.agent_id       = agent_id

        self.trace_id       = None
        self.root_span      = None
        self.root_span_name = None


    def start_root_span(self,
                        root_span_name: str,
                        task_input: Dict[str, Any],
                        trace_id: Optional[str] = None):
        if self.root_span is None:
            trace = None
            if trace_id:
                self.trace_id = trace_id
                trace = XGATaskLangFuse.langfuse.trace(id=trace_id)
            else:
                trace = XGATaskLangFuse.langfuse.trace(name="xga_task_engine", session_id=self.session_id)
                self.trace_id = trace.id

            metadata = {
                'task_id'       : self.task_id,
                'session_id'    : self.session_id,
                'user_id'       : self.user_id,
                'agent_id'      : self.agent_id
            }

            self.root_span = trace.span(
                id          = self.task_run_id,
                name        = f"{root_span_name}[{self.task_no}]",
                input       = task_input,
                metadata    = metadata
            )
            self.root_span_name = root_span_name

            logging.info(f"{root_span_name} TASK_INPUT: {task_input}")

    def end_root_span(self, root_span_name:str, output: Optional[XGATaskResult]=None):
        if self.root_span and self.root_span_name == root_span_name:
            self.root_span.end(output=output)
            self.root_span = None
            self.root_span_name = None


    def create_llm_langfuse_meta(self, llm_count:int)-> LangfuseMetadata:
        generation_name = f"xga_engine_llm_completion[{self.task_no}]({llm_count})"
        generation_id = f"{self.task_run_id}({llm_count})"
        return LangfuseMetadata(
            generation_name     = generation_name,
            generation_id       = generation_id,
            existing_trace_id   = self.trace_id,
            session_id          = self.session_id,
        )